package pp.s1184725.boppi.test;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;
import org.junit.Test;

/**
 * Tests for function use and simple function passing (i.e. no closures required).
 * 
 * @author Frank Wibbelink
 */
public class SimpleFunctionTest {

	/**
	 * Correct function declaration and calls
	 */
	@Test
	public void correctFunctionParsing() {
		AllTests.parseString("function int id(int a) a; 1");
		assertThat(AllTests.log, is(empty()));

		AllTests.parseString("function int id(int a) a; id(1)");
		assertThat(AllTests.log, is(empty()));

		AllTests.parseString("var int a; a := 1; function int const() a; const()");
		assertThat(AllTests.log, is(empty()));

		AllTests.parseString("function int add(int a, int b) a+b; print(add(4, 5)+1)");
		assertThat(AllTests.log, is(empty()));

		AllTests.parseString("function char const(char a, char b) a; print(const('A', 'T'))");
		assertThat(AllTests.log, is(empty()));

		AllTests.parseString("function int factorial(int n) {if (n > 1) then n*factorial(n-1) else 1 fi}");
		assertThat(AllTests.log, is(empty()));
	}

	/**
	 * Function use with syntax errors
	 */
	@Test
	public void wrongFunctionParsing() {
		AllTests.parseString("function doNothing(int a, char b) a+b; print(add(4, 'T')+1)");
		assertThat(AllTests.log, hasSize(1));

		AllTests.parseString("function char add(int a, int b);");
		assertThat(AllTests.log, hasSize(1));

		AllTests.parseString("function int add(int a, int b) a+b; add(4,)");
		assertThat(AllTests.log, hasSize(1));
	}

	/**
	 * Correct function declaration and calls
	 */
	@Test
	public void correctFunctionChecking() {
		AllTests.checkString("function int id(int a) a; 1");
		assertThat(AllTests.log, is(empty()));

		AllTests.checkString("function int id(int a) a; id(1); 1");
		assertThat(AllTests.log, is(empty()));

		AllTests.checkString("var int a; function int const() a; 1");
		assertThat(AllTests.log, is(empty()));

		AllTests.checkString("function int add(int a, int b) a+b; print(add(4, 5)+1)");
		assertThat(AllTests.log, is(empty()));

		AllTests.checkString("function char const(char a, char b) a; print(const('A', 'T'))");
		assertThat(AllTests.log, is(empty()));

		AllTests.checkString("function int factorial(int n) {if (n > 1) then n*factorial(n-1) else 1 fi}");
		assertThat(AllTests.log, is(empty()));
	}

	/**
	 * Function use with type errors
	 */
	@Test
	public void wrongFunctionChecking() {
		AllTests.checkString("function int add(int a, char b) a+b; print(add(4, 'T')+1)");
		assertThat(AllTests.log, hasSize(1));

		AllTests.checkString("function char add(int a, int b) a+b; print(add(4, 5))");
		assertThat(AllTests.log, hasSize(1));

		AllTests.checkString("function int add(int a, int b) a+b; print(add(4, 'T')+1)");
		assertThat(AllTests.log, hasSize(1));
	}

	/**
	 * Correct evaluation of basic functions
	 */
	@Test
	public void correctSimpleFunctionGeneration() {
		AllTests.compileAndRunString("function int id(int a) a; 1");
		assertThat(AllTests.log, is(empty()));

		AllTests.compileAndRunString("function int id(int a) a; id(1); 1");
		assertThat(AllTests.log, is(empty()));

		AllTests.compileAndRunString("var int a; function int const(int b) a; 1");
		assertThat(AllTests.log, is(empty()));

		AllTests.compileAndRunString("var int a; a := 1; function int const(int b) a; print(const(4))");
		assertThat(AllTests.log, is(empty()));
		assertThat(AllTests.out, is(arrayContaining("1")));

		AllTests.compileAndRunString("function int add(int a, int b) a+b; print(add(4, 5)+1)");
		assertThat(AllTests.log, is(empty()));
		assertThat(AllTests.out, is(arrayContaining("10")));

		AllTests.compileAndRunString("function char const(char a, char b) a; print(const('A', 'T'))");
		assertThat(AllTests.log, is(empty()));
		assertThat(AllTests.out, is(arrayContaining("A")));
	}

	/**
	 * Correct evaluation of recursive functions, nested functions and function
	 * passing
	 */
	@Test
	public void correctComplexFunctionGeneration() {
		AllTests.compileAndRunString("function int factorial(int n) {if (n > 1) then n*factorial(n-1) else 1 fi}");
		assertThat(AllTests.log, is(empty()));

		AllTests.compileAndRunFile("recursiveFactorial.boppi", "5");
		assertThat(AllTests.log, is(empty()));
		assertThat(AllTests.out, is(arrayContaining("120")));

		AllTests.compileAndRunFile("simpleFunctionPassing.boppi");
		assertThat(AllTests.log, is(empty()));
		assertThat(AllTests.out, is(arrayContaining("40", "104", "1", "2", "8")));
	}

}
